use anyhow::Result;
use celery::error::TaskError;
use celery::task::TaskResult;
use chrono::prelude::*;
use diesel::prelude::*;
use std::thread;
use std::time::Duration;

use crate::{config, db};

async fn fetch_schoolyears(client: &untis::Client, conn: &mut PgConnection) -> Result<i32> {
    let existing_schoolyears = db::schema::schoolyears::table
        .select(db::schema::schoolyears::untis_id)
        .load::<i32>(conn)?;
    diesel::insert_into(db::schema::schoolyears::table)
        .values(
            &client
                .schoolyears()
                .await?
                .iter()
                .filter(|y| !existing_schoolyears.contains(&y.id))
                .map(|y| db::models::NewSchoolyear {
                    untis_id: y.id,
                    name: &y.name,
                    start_date: y.start_date,
                    end_date: y.end_date,
                })
                .collect::<Vec<db::models::NewSchoolyear>>(),
        )
        .execute(conn)?;

    Ok(db::schema::schoolyears::table
        .filter(db::schema::schoolyears::untis_id.eq(client.current_schoolyear().await?.id))
        .select(db::schema::schoolyears::id)
        .first(conn)?)
}

async fn fetch_current_tenant(
    client: &untis::Client,
    conn: &mut PgConnection,
    schoolyear_id: i32,
) -> Result<()> {
    let tenant = client.current_tenant().await?;
    if diesel::select(diesel::dsl::not(diesel::dsl::exists(
        db::schema::tenants::table.filter(db::schema::tenants::untis_id.eq(tenant.id)),
    )))
    .get_result::<bool>(conn)?
    {
        diesel::update(db::schema::tenants::table)
            .filter(db::schema::tenants::active)
            .set(db::schema::tenants::active.eq(false))
            .execute(conn)?;
        diesel::insert_into(db::schema::tenants::table)
            .values(db::models::NewTenant {
                untis_id: tenant.id,
                schoolyear_id,
                name: &tenant.display_name,
                active: true,
            })
            .execute(conn)?;
    } else if diesel::select(diesel::dsl::exists(
        db::schema::tenants::table
            .filter(db::schema::tenants::untis_id.eq(tenant.id))
            .filter(db::schema::tenants::active.eq(false)),
    ))
    .get_result::<bool>(conn)?
    {
        diesel::update(db::schema::tenants::table)
            .filter(db::schema::tenants::active)
            .set((
                db::schema::tenants::active.eq(false),
                db::schema::tenants::updated_at.eq(diesel::dsl::now),
            ))
            .execute(conn)?;
        diesel::update(db::schema::tenants::table)
            .filter(db::schema::tenants::untis_id.eq(tenant.id))
            .set((
                db::schema::tenants::active.eq(true),
                db::schema::tenants::updated_at.eq(diesel::dsl::now),
            ))
            .execute(conn)?;
    }

    Ok(())
}

async fn fetch_teachers(
    client: &untis::Client,
    conn: &mut PgConnection,
    schoolyear_id: i32,
) -> Result<()> {
    let existing_teachers = db::schema::teachers::table
        .select(db::schema::teachers::untis_id)
        .filter(db::schema::teachers::schoolyear_id.eq(schoolyear_id))
        .load::<i32>(conn)?;
    diesel::insert_into(db::schema::teachers::table)
        .values(
            &client
                .teachers()
                .await?
                .iter()
                .filter(|t| !existing_teachers.contains(&t.id))
                .map(|t| db::models::NewTeacher {
                    untis_id: t.id,
                    schoolyear_id,
                    name: &t.name,
                    forename: if t.forename.is_empty() {
                        None
                    } else {
                        Some(&t.forename)
                    },
                    display_name: &t.display_name,
                })
                .collect::<Vec<db::models::NewTeacher>>(),
        )
        .execute(conn)?;

    Ok(())
}

async fn fetch_classes(
    client: &untis::Client,
    conn: &mut PgConnection,
    schoolyear_id: i32,
) -> Result<()> {
    let existing_classes = db::schema::classes::table
        .select(db::schema::classes::untis_id)
        .filter(db::schema::classes::schoolyear_id.eq(schoolyear_id))
        .load::<i32>(conn)?;
    diesel::insert_into(db::schema::classes::table)
        .values(
            &client
                .classes()
                .await?
                .iter()
                .filter(|c| !existing_classes.contains(&c.id))
                .map(|c| db::models::NewClass {
                    untis_id: c.id,
                    schoolyear_id,
                    name: &c.name,
                    long_name: &c.long_name,
                    active: c.active,
                })
                .collect::<Vec<db::models::NewClass>>(),
        )
        .execute(conn)?;

    Ok(())
}

async fn fetch_subjects(
    client: &untis::Client,
    conn: &mut PgConnection,
    schoolyear_id: i32,
) -> Result<()> {
    let existing_classes = db::schema::subjects::table
        .select(db::schema::subjects::untis_id)
        .filter(db::schema::subjects::schoolyear_id.eq(schoolyear_id))
        .load::<i32>(conn)?;
    diesel::insert_into(db::schema::subjects::table)
        .values(
            &client
                .subjects()
                .await?
                .iter()
                .filter(|c| !existing_classes.contains(&c.id))
                .map(|c| db::models::NewSubject {
                    untis_id: c.id,
                    schoolyear_id,
                    name: &c.name,
                    long_name: &c.long_name,
                })
                .collect::<Vec<db::models::NewSubject>>(),
        )
        .execute(conn)?;

    Ok(())
}

async fn fetch_rooms(
    client: &untis::Client,
    conn: &mut PgConnection,
    schoolyear_id: i32,
) -> Result<()> {
    let existing_classes = db::schema::rooms::table
        .select(db::schema::rooms::untis_id)
        .filter(db::schema::rooms::schoolyear_id.eq(schoolyear_id))
        .load::<i32>(conn)?;
    diesel::insert_into(db::schema::rooms::table)
        .values(
            &client
                .rooms()
                .await?
                .iter()
                .filter(|c| !existing_classes.contains(&c.id))
                .map(|c| db::models::NewRoom {
                    untis_id: c.id,
                    schoolyear_id,
                    name: &c.name,
                    long_name: &c.long_name,
                })
                .collect::<Vec<db::models::NewRoom>>(),
        )
        .execute(conn)?;

    Ok(())
}

async fn fetch_departments(
    client: &untis::Client,
    conn: &mut PgConnection,
    schoolyear_id: i32,
) -> Result<()> {
    let existing_classes = db::schema::departments::table
        .select(db::schema::departments::untis_id)
        .filter(db::schema::departments::schoolyear_id.eq(schoolyear_id))
        .load::<i32>(conn)?;
    diesel::insert_into(db::schema::departments::table)
        .values(
            &client
                .departments()
                .await?
                .iter()
                .filter(|c| !existing_classes.contains(&c.id))
                .map(|c| db::models::NewDepartment {
                    untis_id: c.id,
                    schoolyear_id,
                    name: &c.name,
                    long_name: &c.long_name,
                })
                .collect::<Vec<db::models::NewDepartment>>(),
        )
        .execute(conn)?;

    Ok(())
}

async fn fetch_holidays(
    client: &untis::Client,
    conn: &mut PgConnection,
    schoolyear_id: i32,
) -> Result<()> {
    let existing_classes = db::schema::holidays::table
        .select(db::schema::holidays::untis_id)
        .filter(db::schema::holidays::schoolyear_id.eq(schoolyear_id))
        .load::<i32>(conn)?;
    diesel::insert_into(db::schema::holidays::table)
        .values(
            &client
                .holidays()
                .await?
                .iter()
                .filter(|c| !existing_classes.contains(&c.id))
                .map(|c| db::models::NewHoliday {
                    untis_id: c.id,
                    schoolyear_id,
                    name: &c.name,
                    long_name: &c.long_name,
                    start_date: c.start_date,
                    end_date: c.end_date,
                })
                .collect::<Vec<db::models::NewHoliday>>(),
        )
        .execute(conn)?;

    Ok(())
}

async fn fetch_substitutions(
    client: &untis::Client,
    conn: &mut PgConnection,
    schoolyear_id: i32,
) -> Result<()> {
    let today = Utc::now().date_naive();
    if diesel::select(diesel::dsl::not(diesel::dsl::exists(
        db::schema::substitution_queries::table
            .filter(db::schema::substitution_queries::date.eq(today)),
    )))
    .get_result::<bool>(conn)?
    {
        diesel::insert_into(db::schema::substitution_queries::table)
            .values(db::models::NewSubstitutionQuery {
                schoolyear_id,
                date: today,
                active: true,
            })
            .execute(conn)?;
    }

    for query in db::schema::substitution_queries::table
        .filter(db::schema::substitution_queries::active)
        .load::<db::models::SubstitutionQuery>(conn)?
    {
        let query_result_id = diesel::insert_into(db::schema::substitution_query_results::table)
            .values(db::models::NewSubstitutionQueryResult {
                substitution_query_id: query.id,
                queried_at: Utc::now().naive_utc(),
            })
            .returning(db::schema::substitution_query_results::id)
            .get_result::<i32>(conn)?;

        let substs = client.substitutions(&query.date, &query.date, None).await?;
        for substitution in substs {
            let substitution_id = diesel::insert_into(db::schema::substitutions::table)
                .values(db::models::NewSubstitution {
                    substitution_query_result_id: query_result_id,
                    subst_type: substitution.subst_type.into(),
                    lesson_id: substitution.lesson_id,
                    start_time: substitution.start_time,
                    end_time: substitution.end_time,
                    text: substitution.text.as_deref(),
                })
                .returning(db::schema::substitutions::id)
                .get_result::<i32>(conn)?;

            diesel::insert_into(db::schema::substitution_classes::table)
                .values(
                    &substitution
                        .classes
                        .iter()
                        .enumerate()
                        .map(|(i, c)| {
                            Ok(db::models::NewSubstitutionClass {
                                substitution_id,
                                position: i as i16,
                                class_id: db::schema::classes::table
                                    .filter(db::schema::classes::untis_id.eq(c.id))
                                    .select(db::schema::classes::id)
                                    .get_result::<i32>(conn)?,
                            })
                        })
                        .collect::<Result<Vec<db::models::NewSubstitutionClass>>>()?,
                )
                .execute(conn)?;
            diesel::insert_into(db::schema::substitution_teachers::table)
                .values(
                    &substitution
                        .teachers
                        .iter()
                        .enumerate()
                        .map(|(i, t)| {
                            Ok(db::models::NewSubstitutionTeacher {
                                substitution_id,
                                position: i as i16,
                                teacher_id: if t.id == 0 {
                                    None
                                } else {
                                    Some(
                                        db::schema::teachers::table
                                            .filter(db::schema::teachers::untis_id.eq(t.id))
                                            .select(db::schema::teachers::id)
                                            .get_result::<i32>(conn)?,
                                    )
                                },
                            })
                        })
                        .collect::<Result<Vec<db::models::NewSubstitutionTeacher>>>()?,
                )
                .execute(conn)?;
            diesel::insert_into(db::schema::substitution_subjects::table)
                .values(
                    &substitution
                        .subjects
                        .iter()
                        .enumerate()
                        .map(|(i, s)| {
                            Ok(db::models::NewSubstitutionSubject {
                                substitution_id,
                                position: i as i16,
                                subject_id: db::schema::subjects::table
                                    .filter(db::schema::subjects::untis_id.eq(s.id))
                                    .select(db::schema::subjects::id)
                                    .get_result::<i32>(conn)?,
                            })
                        })
                        .collect::<Result<Vec<db::models::NewSubstitutionSubject>>>()?,
                )
                .execute(conn)?;
            diesel::insert_into(db::schema::substitution_rooms::table)
                .values(
                    &substitution
                        .rooms
                        .iter()
                        .enumerate()
                        .map(|(i, r)| {
                            Ok(db::models::NewSubstitutionRoom {
                                substitution_id,
                                position: i as i16,
                                room_id: if r.id == 0 {
                                    None
                                } else {
                                    Some(
                                        db::schema::rooms::table
                                            .filter(db::schema::rooms::untis_id.eq(r.id))
                                            .select(db::schema::rooms::id)
                                            .get_result::<i32>(conn)?,
                                    )
                                },
                            })
                        })
                        .collect::<Result<Vec<db::models::NewSubstitutionRoom>>>()?,
                )
                .execute(conn)?;
        }
    }

    Ok(())
}

#[celery::task]
pub async fn update_info() -> TaskResult<()> {
    let dur = Duration::from_secs(10);

    thread::sleep(dur);
    dbg!("DONE!");
    let mut client = match config::untis_from_env() {
        Ok(x) => x,
        Err(e) => return Err(TaskError::UnexpectedError(e.to_string())),
    };
    if let Err(e) = client.login().await {
        return Err(TaskError::UnexpectedError(e.to_string()));
    }
    thread::sleep(dur);
    dbg!("DONE!");

    let conn = &mut match db::POOL.get() {
        Ok(x) => x,
        Err(e) => return Err(TaskError::UnexpectedError(e.to_string())),
    };

    let schoolyear_id = match fetch_schoolyears(&client, conn).await {
        Ok(x) => x,
        Err(e) => return Err(TaskError::UnexpectedError(e.to_string())),
    };
    thread::sleep(dur);
    dbg!("DONE!");
    if let Err(e) = fetch_current_tenant(&client, conn, schoolyear_id).await {
        return Err(TaskError::UnexpectedError(e.to_string()));
    }
    thread::sleep(dur);
    dbg!("DONE!");
    if let Err(e) = fetch_teachers(&client, conn, schoolyear_id).await {
        return Err(TaskError::UnexpectedError(e.to_string()));
    }
    thread::sleep(dur);
    dbg!("DONE!");
    if let Err(e) = fetch_classes(&client, conn, schoolyear_id).await {
        return Err(TaskError::UnexpectedError(e.to_string()));
    }
    thread::sleep(dur);
    dbg!("DONE!");
    if let Err(e) = fetch_subjects(&client, conn, schoolyear_id).await {
        return Err(TaskError::UnexpectedError(e.to_string()));
    }
    thread::sleep(dur);
    dbg!("DONE!");
    if let Err(e) = fetch_rooms(&client, conn, schoolyear_id).await {
        return Err(TaskError::UnexpectedError(e.to_string()));
    }
    thread::sleep(dur);
    dbg!("DONE!");
    if let Err(e) = fetch_departments(&client, conn, schoolyear_id).await {
        return Err(TaskError::UnexpectedError(e.to_string()));
    }
    thread::sleep(dur);
    dbg!("DONE!");
    if let Err(e) = fetch_holidays(&client, conn, schoolyear_id).await {
        return Err(TaskError::UnexpectedError(e.to_string()));
    }
    thread::sleep(dur);
    dbg!("DONE!");
    if let Err(e) = fetch_substitutions(&client, conn, schoolyear_id).await {
        return Err(TaskError::UnexpectedError(e.to_string()));
    }
    thread::sleep(dur);
    dbg!("DONE!");

    if let Err(e) = client.logout().await {
        return Err(TaskError::UnexpectedError(e.to_string()));
    }
    thread::sleep(dur);
    dbg!("DONE!");

    Ok(())
}
